"""Docker functions to get info about containers"""

from docker import APIClient, client as DockerClient, utils, errors
from requests.exceptions import ConnectionError
cts_info = dict()
running_cts = 0

params = utils.kwargs_from_env()
base_url = None if 'base_url' not in params else params['base_url']
tls = None if 'tls' not in params else params['tls']
apiclient = APIClient(base_url=base_url, tls=tls)
client = DockerClient.from_env()


def block_ct_ports(service: str, ports: list, project_name: str) -> tuple:
    try:
        ct = client.containers.get(get_ct_item(service, 'id'))
    except Exception:
        return (False, '{} is not started, no port to block'.format(service))

    iptables = ct.exec_run(['which', 'iptables']).decode().strip()
    if iptables == '':
        return (True, "Can't block ports on {}, is iptables installed ?".format(service))

    _allow_contact_subnet(project_name, ct)

    # Now for each port, add an iptable rule
    for port in ports:
        try:
            ct.exec_run([iptables, '-D', 'OUTPUT', '-p', 'tcp', '--dport', port, '-j', 'REJECT'])
        finally:
            ct.exec_run([iptables, '-A', 'OUTPUT', '-p', 'tcp', '--dport', port, '-j', 'REJECT'])

    return (False, 'Blocked ports {} on container {}'.format(', '.join(ports), service))


def check_cts_are_running(project_name: str, config: str = None):
    """Throws an error if cts are not running"""

    global cts_info

    running_cts, cts_info = get_running_containers(project_name, config)
    if running_cts is 0:
        raise SystemError('Have you started your server with the start action ?')

    return (running_cts, cts_info)


def container_running(container: str):
    """Returns True if the container is running else False"""

    try:
        ct_data = apiclient.inspect_container(container)

        return ct_data['State']['Running']
    except Exception:
        return False


def get_ct_item(compose_name: str, item_name: str):
    """Get a value from a container, such as name or IP"""

    if len(cts_info) is 0:
        raise RuntimeError('Before getting an info from a ct, run check_cts_are_running()')

    for ct_id, ct_data in cts_info.items():
        if ct_data['compose_name'] == compose_name:
            return ct_data[item_name]

    return ''


def get_ct_name(container: str):
    ct_name = get_ct_item(container, 'name')
    if ct_name == '':
        raise Exception('{} does not seem to be started ...'.format(container))

    return ct_name


def get_running_containers(project_name: str, config: str = None):
    """Get a list of IDs of running containers for the current stakkr instance"""

    filters = {
        'name': '{}_'.format(project_name),
        'status': 'running',
        'network': '{}_stakkr'.format(project_name).replace('-', '')}

    try:
        cts = client.containers.list(filters=filters)
    except ConnectionError:
        raise Exception('Make sure docker is installed and running')

    for ct in cts:
        container_info = _extract_container_info(project_name, ct.id)
        cts_info[container_info['name']] = container_info

    return (len(cts), cts_info)


def _extract_container_info(project_name: str, ct_id: str):
    """Get a hash of info about a container : name, ports, image, ip ..."""

    try:
        ct_data = apiclient.inspect_container(ct_id)
        ct_info = {
            'id': ct_id,
            'name': ct_data['Name'].lstrip('/'),
            'compose_name': ct_data['Config']['Labels']['com.docker.compose.service'],
            'ports': ct_data['Config']['ExposedPorts'].keys() if 'ExposedPorts' in ct_data['Config'] else [],
            'image': ct_data['Config']['Image'],
            'ip': _get_ip_from_networks(project_name, ct_data['NetworkSettings']['Networks']),
            'running': ct_data['State']['Running']
        }

        return ct_info
    except Exception as e:
        return None


def _get_ip_from_networks(project_name: str, networks: list):
    """Get a list of IPs for a network"""

    project_name = project_name.replace('-', '')
    network_settings = {}
    if '{}_stakkr'.format(project_name) in networks:
        network_settings = networks['{}_stakkr'.format(project_name)]

    return network_settings['IPAddress'] if 'IPAddress' in network_settings else ''


def create_network(network: str):
    """Create a Network"""

    if network_exists(network):
        return False

    return client.networks.create(network, driver='bridge').id


def get_subnet(project_name: str):
    network_info = client.networks.get(project_name.replace('-', '') + '_stakkr').attrs

    return network_info['IPAM']['Config'][0]['Subnet'].split('/')[0]


def get_switch_ip():
    import socket

    cmd = r"""/bin/sh -c "ip addr show hvint0 | grep -oE '([0-9]{1,3}\.){3}[0-9]{1,3}'" """
    res = client.containers.run(
        'alpine', remove=True, tty=True, privileged=True, network_mode='host', pid_mode='host', command=cmd)
    ip = res.strip().decode()

    try:
        socket.inet_aton(ip)
        return ip
    except socket.error:
        raise ValueError('{} is not a valid ip, check docker is running')


def network_exists(network: str):
    try:
        client.networks.get(network)
        return True
    except errors.NotFound:
        return False


def add_container_to_network(container: str, network: str):
    """Attach a container to a network"""

    if _container_in_network(container, network) is True:
        return False

    docker_network = client.networks.get(network)
    docker_network.connect(container)

    return True


def _allow_contact_subnet(project_name: str, ct) -> None:
    iptables = ct.exec_run(['which', 'iptables']).decode().strip()
    if iptables == '':
        return False

    subnet = get_subnet(project_name) + '/24'
    # Allow internal network
    try:
        ct.exec_run([iptables, '-D', 'OUTPUT', '-d', subnet, '-j', 'ACCEPT'])
    finally:
        ct.exec_run([iptables, '-A', 'OUTPUT', '-d', subnet, '-j', 'ACCEPT'])


def _container_in_network(container: str, expected_network: str):
    """Returns True if a container is in a network else false. Used by add_container_to_network"""

    try:
        ct_data = apiclient.inspect_container(container)
    except Exception:
        raise SystemError('Container {} does not seem to exist'.format(container))

    for connected_network in ct_data['NetworkSettings']['Networks'].keys():
        if connected_network == expected_network:
            return True

    return False



import click
import os
import sys

from stakkr import package_utils
from stakkr.configreader import Config
from subprocess import Popen


@click.command(help="Wrapper for docker-compose", context_settings=dict(ignore_unknown_options=True))
@click.option('--config', '-c', help="Override the conf/compose.ini")
@click.argument('command', nargs=-1, type=click.UNPROCESSED)
def cli(config: str, command):
    config = Config(config)
    main_config = config.read()
    if main_config is False:
        config.display_errors()
        sys.exit(1)

    main_config = main_config['main']
    set_env_values_from_conf(main_config)

    project_name = main_config.get('project_name')
    os.putenv('COMPOSE_PROJECT_NAME', project_name)

    # Services from config
    configured_services = main_config.get('services')

    # Services available from base and plugins
    available_services = get_available_services()
    available_services = add_services_from_plugins(available_services)
    available_services = add_local_services(available_services)
    services = get_enabled_services(configured_services, available_services)

    cmd = ['docker-compose', '-f', package_utils.get_file('static', 'docker-compose.yml')] + services
    cmd += ['-p', project_name] + list(command)

    Popen(cmd)


def add_services_from_plugins(available_services: list):
    from pkg_resources import iter_entry_points

    # Override services with plugins
    for entry in iter_entry_points('stakkr.plugins'):
        plugin_dir = str(entry).split('=')[0].strip()
        services_dir = package_utils.get_venv_basedir() + '/plugins/' + plugin_dir + '/services'

        conf_files = _get_services_from_dir(services_dir)
        if conf_files is []:
            continue

        for conf_file in conf_files:
            available_services[conf_file[:-4]] = services_dir + '/' + conf_file

    return available_services


def add_local_services(available_services: list):
    services_dir = package_utils.get_venv_basedir() + '/services'

    conf_files = _get_services_from_dir(services_dir)
    for conf_file in conf_files:
        available_services[conf_file[:-4]] = services_dir + '/' + conf_file

    return available_services


def get_available_services():
    services_dir = package_utils.get_dir('static') + '/services/'
    conf_files = _get_services_from_dir(services_dir)

    services = dict()
    for conf_file in conf_files:
        services[conf_file[:-4]] = services_dir + conf_file

    return services


def get_enabled_services(configured_services: list, available_services: list):
    services_files = []
    for service in configured_services:
        if service not in available_services:
            print(click.style('Error: service "{}" has no configuration file. Check your compose.ini'.format(service), fg='red'))
            sys.exit(1)
        services_files.append('-f')
        services_files.append(available_services[service])

    return services_files


def set_env_values_from_conf(config: list):
    os.environ['DOCKER_UID'] = _get_uid(config.pop('uid'))
    os.environ['DOCKER_GID'] = _get_gid(config.pop('gid'))
    os.environ['COMPOSE_BASE_DIR'] = package_utils.get_venv_basedir()

    for parameter, value in config.items():
        parameter = 'DOCKER_{}'.format(parameter.replace('.', '_').upper())
        os.environ[parameter] = str(value)


def _get_services_from_dir(services_dir: str):
    if os.path.isdir(services_dir) is False:
        return []

    return [service for service in os.listdir(services_dir) if service.endswith('.yml')]


def _get_uid(uid: int):
    if uid is not None:
        return str(uid)

    return '1000' if os.name == 'nt' else str(os.getuid())


def _get_gid(gid: int):
    if gid is not None:
        return str(gid)

    return '1000' if os.name == 'nt' else str(os.getgid())


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
"""Simple Config Reader"""

import os
import sys

from configobj import ConfigObj, flatten_errors
from stakkr import package_utils
from validate import Validator


class Config():
    """Parser of Stakkr. Set default values and validate
    conf/compose.ini with conf/configspec.ini

    """

    def __init__(self, config_file: str = None):
        """If no config given, read the default one"""
        self.errors = dict()
        self.config_file = package_utils.get_venv_basedir() + '/conf/compose.ini'
        if config_file is not None:
            self.config_file = config_file


    def display_errors(self):
        """Display errors in STDOUT"""
        from click import style

        print(style('Failed validating {}: '.format(self.config_file), fg='red'), file=sys.stderr)
        for key, error in self.errors.items():
            print('  - "{}" : {}'.format(style(key, fg='yellow'), error), file=sys.stderr)


    def read(self):
        """Read the default values and overriden ones"""
        if os.path.isfile(self.config_file) is False:
            raise IOError('Config file {} does not exist'.format(self.config_file))

        return self._parse()


    def _parse(self):
        """Parse the config from configspecs that is a file either local or from a package"""

        config = ConfigObj(infile=self.config_file, configspec=package_utils.get_file('static', 'configspec.ini'))

        validator = Validator()
        validated = config.validate(validator, preserve_errors=True)
        if validated is not True:
            self._register_errors(config, validated)
            return False

        return config


    def _register_errors(self, config: dict, validated):
        for [sectionList, key, error] in flatten_errors(config, validated):
            if key is not None:
                error = 'Missing' if error is False else str(error)
                self.errors[key] = error



"""
A command wrapper to get a live output displayed. Useful when you need to write a plugin that outputs some progress
bar or info.
"""

import subprocess
import sys

from click import style
from io import BufferedReader


def launch_cmd_displays_output(cmd: list, print_msg=True, print_err=True, err_to_out=False):
    """Launch a command and displays conditionnaly messages and / or errors"""

    try:
        stderr = subprocess.PIPE if err_to_out is False else subprocess.STDOUT
        result = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=stderr)
    except Exception as e:
        raise SystemError('Cannot run the command: {}'.format(e))

    _read_messages(result, print_msg)
    if print_err is True and err_to_out is False:
        _print_errors(result)

    return result


def _read_messages(result: BufferedReader, display: bool = False):
    """Print messages sent to the STDOUT"""

    for line in result.stdout:
        line = line.decode()
        line = line if display is True else '.'
        print(line, end='')
        sys.stdout.flush()

    print()


def _print_errors(result: BufferedReader):
    """Print messages sent to the STDERR"""

    i = 0
    for line in result.stderr:
        err = line.decode()

        if i == 0:
            print(style("Command returned errors :", fg='red'))

        if i < 5:
            print(err, end='')
        elif i is 5:
            print(style('... and more', fg='red'))

        i += 1



import click
import sys

from . import package_utils
from click_plugins import with_plugins
from pkg_resources import iter_entry_points


@with_plugins(iter_entry_points('stakkr.plugins'))
@click.group(help="""Main CLI Tool that easily create / maintain
a stack of services, for example for web development.

Read the configuration file and setup the required services by
linking and managing everything for you.""")
@click.version_option('3.3.2')
@click.option('--config', '-c', help='Change the configuration file')
@click.option('--debug/--no-debug', '-d', default=False)
@click.option('--verbose', '-v', is_flag=True)
@click.pass_context
def stakkr(ctx, config, debug, verbose):
    from stakkr.actions import StakkrActions

    # Add the virtual env in the path
    venv_base = package_utils.get_venv_basedir()
    sys.path.append(venv_base)

    ctx.obj['CONFIG'] = config
    ctx.obj['DEBUG'] = debug
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['STAKKR'] = StakkrActions(venv_base, ctx.obj)


@stakkr.command(help="""Enter a container to perform direct actions such as install packages, run commands ...

Valid values for CONTAINER : 'apache', 'mysql' or 'php'""")
@click.option('--user', '-u', help="User's name. Valid choices : www-data or root", default='www-data')
@click.argument('container', required=True)
@click.pass_context
def console(ctx, container: str, user: str):
    cmd_user = None

    if container not in ['php', 'apache'] and user == 'www-data':
        cmd_user = 'root'
    elif cmd_user is None:
        cmd_user = user

    ctx.obj['STAKKR'].console(container, cmd_user)


@stakkr.command(
    help="""Start or Stop the DNS forwarder.
Only one DNS Forwarded by host is possible. Done with mgood/resolvable.
Also, that is completely useless under Windows as we can't mount /etc/resolv.conf

Valid values for ACTION : 'start' or 'stop'""",
    name="dns"
    )
@click.argument('action', required=True, type=click.Choice(['start', 'stop']))
@click.pass_context
def dns(ctx, action: str):
    print(click.style('[{}]'.format(action.upper()), fg='green') + ' DNS forwarder ...')
    ctx.obj['STAKKR'].manage_dns(action)
    _show_status(ctx)


@stakkr.command(help="""Execute a command into a container.

Examples:\n
- ``stakkr -v exec mysql mysqldump -p'$MYSQL_ROOT_PASSWORD' mydb > /tmp/backup.sql``\n
- ``stakkr exec php php -v`` : Execute the php binary in the php container with option -v\n
- ``stakkr exec apache service apache2 restart``\n
""", context_settings=dict(ignore_unknown_options=True))
@click.pass_context
@click.option('--user', '-u', help="User's name. Be careful, each container have its own users.")
@click.argument('container', required=True)
@click.argument('command', required=True, nargs=-1, type=click.UNPROCESSED)
def exec(ctx, user: str, container: str, command: tuple):
    users = {
        'php': 'www-data'
    }
    if user is None:
        user = users[container] if container in users else 'root'

    ctx.obj['STAKKR'].exec(container, user, command)


@stakkr.command(
    help="""`stakkr mysql` is a wrapper for the mysql binary located in the mysql service.

You can run any mysql command as root, such as :\n
- ``stakkr mysql -e "CREATE DATABASE mydb"`` to create a DB from outside\n
- ``stakkr mysql`` to enter the mysql console\n
- ``cat myfile.sql | stakkr mysql mydb`` to import a file from outside to mysql\n

For scripts, you must use the relative path.
""", context_settings=dict(ignore_unknown_options=True))
@click.pass_context
@click.argument('command', nargs=-1, type=click.UNPROCESSED)
def mysql(ctx, command: tuple):
    command = ('mysql', '-p$MYSQL_ROOT_PASSWORD') + command
    ctx.invoke(exec, user='root', container='mysql', command=command)


@stakkr.command(help='Required to be launched if you install a new plugin', name="refresh-plugins")
@click.pass_context
def refresh_plugins(ctx):
    from stakkr.plugins import add_plugins

    print(click.style('Adding plugins from plugins/', fg='green'))
    plugins = add_plugins()
    if len(plugins) is 0:
        print(click.style('No plugin to add', fg='yellow'))
        exit(0)

    print()
    print(click.style('Plugins refreshed.\n', fg='green'))


@stakkr.command(help="Restart all containers")
@click.option('--pull', '-p', help="Force a pull of the latest images versions", is_flag=True)
@click.option('--recreate', '-r', help="Recreate all containers", is_flag=True)
@click.pass_context
def restart(ctx, pull: bool, recreate: bool):
    print(click.style('[RESTARTING]', fg='green') + ' your stakkr services')
    try:
        ctx.invoke(stop)
    except Exception:
        print()

    ctx.invoke(start, pull=pull, recreate=recreate)


@stakkr.command(help='DEPRECATED, USE `stakkr exec`', context_settings=dict(ignore_unknown_options=True))
@click.pass_context
@click.option('--user', '-u', help="User's name. Valid values: www-data or root", type=click.Choice(['www-data', 'root']))
@click.argument('container', required=True, type=click.Choice(['mysql', 'php']))
@click.argument('run_args', nargs=-1, type=click.UNPROCESSED)
def run(ctx, container: str, user: str, run_args: tuple):
    print(click.style('[DEPRECATED]', fg='red') + ' You must use either `stakkr mysql` or `stakkr exec php php`')


@stakkr.command(help="Start containers defined in compose.ini")
@click.option('--pull', '-p', help="Force a pull of the latest images versions", is_flag=True)
@click.option('--recreate', '-r', help="Recreate all containers", is_flag=True)
@click.pass_context
def start(ctx, pull: bool, recreate: bool):
    print(click.style('[STARTING]', fg='green') + ' your stakkr services')
    ctx.obj['STAKKR'].start(pull, recreate)
    _show_status(ctx)


@stakkr.command(help="Display a list of running containers")
@click.pass_context
def status(ctx):
    ctx.obj['STAKKR'].status()


@stakkr.command(help="Stop the services")
@click.pass_context
def stop(ctx):
    print(click.style('[STOPPING]', fg='yellow') + ' your stakkr services')
    ctx.obj['STAKKR'].stop()


def _show_status(ctx):
    services_ports = ctx.obj['STAKKR'].get_services_ports()
    if services_ports == '':
        print('\nServices Status:')
        ctx.invoke(status)
        return

    print('\nServices URLs :')
    print(services_ports)


def debug_mode():
    if '--debug' in sys.argv or '-d' in sys.argv:
        return True

    return False


def main():
    try:
        stakkr(obj={})
    except Exception as e:
        msg = click.style(r""" ______ _____  _____   ____  _____
|  ____|  __ \|  __ \ / __ \|  __ \
| |__  | |__) | |__) | |  | | |__) |
|  __| |  _  /|  _  /| |  | |  _  /
| |____| | \ \| | \ \| |__| | | \ \
|______|_|  \_\_|  \_\\____/|_|  \_\

""", fg='yellow')
        msg += click.style('{}'.format(e), fg='red')
        print(msg + '\n', file=sys.stderr)

        if debug_mode() is True:
            raise e

        sys.exit(1)


if __name__ == '__main__':
    main()



"""Module used by setup.py to find plugins to load with click"""

import re
import subprocess

from os import listdir, path


def add_plugins():
    """Read the plugins directory, get the subfolders from it and look for .py files"""

    if path.isdir('plugins') is False:
        return []

    _remove_plugins()
    folders = _get_subfolders('plugins')

    plugins = []
    for folder in folders:
        plugins = _add_plugin_from_dir(plugins, 'plugins/{}/'.format(folder))

    return sorted(plugins)


def _add_plugin_from_dir(plugins: list, full_path: str):
    files = _get_files_from_folder(full_path)
    for plugin in files:
        plugin_name = full_path.strip('/').split('/')[1]
        try:
            subprocess.check_call(['pip', 'install', '--upgrade', full_path], stdout=subprocess.DEVNULL)
            print('  -> Plugin "{}" added'.format(plugin_name))
            plugins.append(plugin_name)

        except Exception as e:
            raise TypeError('Problem installing {} (Reason: {})'.format(plugin_name[:-3], e))

    return plugins


def _get_files_from_folder(full_path: str):
    files = listdir(full_path)

    return [filename for filename in files if filename == 'setup.py']


def _get_subfolders(directory: str):
    subfolders = listdir(directory)

    return [folder for folder in subfolders if path.isdir('{}/{}'.format(directory, folder))]


def _remove_plugins():
    cmd = ['pip', 'freeze']
    res = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    regex = re.compile('^stakkr(.+?)==(.+)$', re.IGNORECASE)

    for line in res.stdout:
        plugin = re.search(regex, line.decode())
        if plugin is None:
            continue

        plugin_name = 'Stakkr{}'.format(plugin.group(1))
        print('  -> Cleaning "{}"'.format(plugin_name))
        subprocess.check_call(['pip', 'uninstall', '-y', plugin_name], stdout=subprocess.DEVNULL)



import click
import os
import sys
import subprocess

from . import command, docker, package_utils
from .configreader import Config
from clint.textui import colored, puts, columns
from docker import client as DockerClient


class StakkrActions():
    """Main class that does actions asked in the cli"""

    _services_to_display = {
        'apache': {'name': 'Web Server', 'url': 'http://{}'},
        'mailcatcher': {'name': 'Mailcatcher (fake SMTP)', 'url': 'http://{}', 'extra_port': 25},
        'maildev': {'name': 'Maildev (Fake SMTP)', 'url': 'http://{}', 'extra_port': 25},
        'phpmyadmin': {'name': 'PhpMyAdmin', 'url': 'http://{}'},
        'xhgui': {'name': 'XHGui (PHP Profiling)', 'url': 'http://{}'}
    }


    def __init__(self, base_dir: str, ctx: dict):
        # Work with directories and move to the right place
        self.stakkr_base_dir = base_dir
        self.context = ctx
        self.current_dir = os.getcwd()
        self.current_dir_relative = self._get_relative_dir()
        os.chdir(self.stakkr_base_dir)

        # Set some general variables
        self.dns_container_name = 'docker_dns'
        self.config_file = ctx['CONFIG']
        self.compose_base_cmd = self._get_compose_base_cmd()
        self.docker_client = DockerClient.from_env()

        # Get info from config
        self.config = self._get_config()
        self.main_config = self.config['main']
        self.project_name = self.main_config.get('project_name')


    def console(self, ct: str, user: str):
        """Enter a container (stakkr allows only apache / php and mysql)"""

        docker.check_cts_are_running(self.project_name, self.config_file)

        cmd = ['docker', 'exec', '-u', user, '-it', docker.get_ct_name(ct), 'bash']
        subprocess.call(cmd)
        self._verbose('Command : "' + ' '.join(cmd) + '"')


    def get_services_ports(self):
        """Once started, stakkr displays a message with the list of launched containers."""

        dns_started = docker.container_running('docker_dns')

        running_cts, cts = docker.get_running_containers(self.project_name, self.config_file)

        text = ''
        for ct, ct_info in cts.items():
            if ct_info['compose_name'] not in self._services_to_display:
                continue
            options = self._services_to_display[ct_info['compose_name']]
            url = self._get_url(options['url'], ct_info['compose_name'], dns_started)
            text += '  - For {}'.format(colored.yellow(options['name'])).ljust(55, ' ') + ' : ' + url + '\n'

            if 'extra_port' in options:
                text += ' '*3 + ' ... in your containers use the port {}\n'.format(options['extra_port'])

        return text


    def exec(self, container: str, user: str, args: tuple):
        """Run a command from outside to any container. Wrapped into /bin/sh"""

        docker.check_cts_are_running(self.project_name, self.config_file)

        args = ['"{}"'.format(arg) for arg in args]

        tty = 't' if sys.stdin.isatty() else ''
        cmd = ['docker', 'exec', '-u', user, '-i' + tty, docker.get_ct_name(container), 'sh', '-c']
        cmd += ["""test -d "/var/{0}" && cd "/var/{0}" ; exec {1}""".format(self.current_dir_relative, ' '.join(args))]
        self._verbose('Command : "' + ' '.join(cmd) + '"')
        subprocess.call(cmd, stdin=sys.stdin)


    def manage_dns(self, action: str):
        """Starts or stop the DNS forwarder"""

        dns_started = docker.container_running(self.dns_container_name)
        if dns_started is True and action == 'stop':
            dns = self.docker_client.containers.get('docker_dns')
            return dns.stop()

        elif dns_started is False and action == 'start':
            self._docker_run_dns()
            return

        puts(colored.red('[ERROR]') + " Can't {} the dns container (already done?)".format(action))
        sys.exit(1)


    def start(self, pull: bool, recreate: bool):
        """If not started, start the containers defined in config"""

        verb = self.context['VERBOSE']
        debug = self.context['DEBUG']

        try:
            docker.check_cts_are_running(self.project_name, self.config_file)
            puts(colored.yellow('[INFO]') + ' stakkr is already started ...')
            sys.exit(0)
        except Exception:
            pass

        if pull is True:
            command.launch_cmd_displays_output(self.compose_base_cmd + ['pull'], verb, debug, True)

        recreate_param = '--force-recreate' if recreate is True else '--no-recreate'
        cmd = self.compose_base_cmd + ['up', '-d', recreate_param, '--remove-orphans']
        self._verbose('Command: ' + ' '.join(cmd))
        command.launch_cmd_displays_output(cmd, verb, debug, True)

        self.running_cts, self.cts = docker.get_running_containers(self.project_name, self.config_file)
        if self.running_cts is 0:
            raise SystemError("Couldn't start the containers, run the start with '-v' and '-d'")

        self._patch_oses_start()
        self._run_iptables_rules()
        self._run_services_post_scripts()


    def status(self):
        """Returns a nice table with the list of started containers"""

        try:
            self.running_cts, self.cts = docker.check_cts_are_running(self.project_name, self.config_file)
        except Exception:
            puts(colored.yellow('[INFO]') + ' stakkr is currently stopped')
            sys.exit(0)

        dns_started = docker.container_running('docker_dns')

        self._print_status_headers(dns_started)
        self._print_status_body(dns_started)


    def stop(self):
        """If started, stop the containers defined in config. Else throw an error"""

        verb = self.context['VERBOSE']
        debug = self.context['DEBUG']

        docker.check_cts_are_running(self.project_name, self.config_file)
        command.launch_cmd_displays_output(self.compose_base_cmd + ['stop'], verb, debug, True)
        self._patch_oses_stop()

        self.running_cts, self.cts = docker.get_running_containers(self.project_name, self.config_file)
        if self.running_cts is not 0:
            raise SystemError("Couldn't stop services ...")


    def _call_service_post_script(self, service: str):
        service_script = package_utils.get_file('static', 'services/{}.sh'.format(service))
        if os.path.isfile(service_script) is True:
            cmd = ['bash', service_script, docker.get_ct_item(service, 'name')]
            subprocess.call(cmd)
            self._verbose('Service Script : ' + ' '.join(cmd))


    def _docker_run_dns(self):
        docker.check_cts_are_running(self.project_name, self.config_file)

        try:
            cmd = ['docker', 'run', '--rm', '-d', '--hostname', 'docker-dns', '--name', self.dns_container_name]
            cmd += ['--network', self.project_name + '_stakkr']
            cmd += ['-v', '/var/run/docker.sock:/tmp/docker.sock', '-v', '/etc/resolv.conf:/tmp/resolv.conf']
            cmd += ['mgood/resolvable']
            self._verbose('Command : ' + ' '.join(cmd))
            command.launch_cmd_displays_output(cmd, self.context['VERBOSE'], self.context['DEBUG'])
        except Exception as e:
            puts(colored.red('[ERROR]') + " Can't start the DNS, maybe it exists already ?")
            print('       -> {}'.format(e))
            sys.exit(1)


    def _get_compose_base_cmd(self):
        if self.context['CONFIG'] is None:
            return ['stakkr-compose']

        return ['stakkr-compose', '-c', self.context['CONFIG']]


    def _get_config(self):
        config = Config(self.config_file)
        main_config = config.read()
        if main_config is False:
            config.display_errors()
            sys.exit(1)

        return main_config


    def _get_relative_dir(self):
        if self.current_dir.startswith(self.stakkr_base_dir):
            return self.current_dir[len(self.stakkr_base_dir):].lstrip('/')

        return ''


    def _get_url(self, service_url: str, service: str, dns_started: bool):
        name = docker.get_ct_name(service) if dns_started else docker.get_ct_item(service, 'ip')

        return service_url.format(name)


    def _patch_oses_start(self):
        """For Windows we need to do a few manipulations to be able to access
        our dev environment from the host ...

        Found here: https://github.com/docker/for-win/issues/221#issuecomment-270576243"""

        if os.name not in ['nt']:
            return

        self.docker_client.containers.run(
            'justincormack/nsenter1', remove=True, tty=True, privileged=True, network_mode='none',
            pid_mode='host', command='bin/sh -c "iptables -A FORWARD -j ACCEPT"')

        subnet = docker.get_subnet(self.project_name)
        switch_ip = docker.get_switch_ip()
        msg = 'We need to create a route for the network {} via {} to work...'
        click.secho(msg.format(subnet, switch_ip), fg='yellow', nl=False)
        subprocess.call(['route', 'add', subnet, 'MASK', '255.255.255.0', switch_ip])
        print()


    def _patch_oses_stop(self):
        """Opposite actions than _patch_oses_start()"""

        if os.name not in ['nt']:
            return

        subnet = docker.get_subnet(self.project_name)
        click.secho("Let's remove the route that has been added for {}...".format(subnet), fg='yellow', nl=False)
        subprocess.call(['route', 'delete', subnet])
        print()


    def _print_status_headers(self, dns_started: bool):
        host_ip = (colored.green('HostName' if dns_started else 'IP'))

        puts(columns(
            [(colored.green('Container')), 16], [host_ip, 25], [(colored.green('Ports')), 25],
            [(colored.green('Image')), 32],
            [(colored.green('Docker ID')), 15], [(colored.green('Docker Name')), 25]
            ))

        puts(columns(
            ['-'*16, 16], ['-'*25, 25], ['-'*25, 25],
            ['-'*32, 32],
            ['-'*15, 15], ['-'*25, 25]
            ))


    def _print_status_body(self, dns_started: bool):
        for ct in sorted(self.cts.keys()):
            ct_data = self.cts[ct]
            if ct_data['ip'] == '':
                continue

            host_ip = ct_data['name'] if dns_started else ct_data['ip']

            puts(columns(
                [ct_data['compose_name'], 16], [host_ip, 25], [', '.join(ct_data['ports']), 25],
                [ct_data['image'], 32],
                [ct_data['id'][:12], 15], [ct_data['name'], 25]
                ))


    def _run_iptables_rules(self):
        """For some containers we need to add iptables rules added from the config"""

        block_config = self.config['network-block']
        for service, ports in block_config.items():
            error, msg = docker.block_ct_ports(service, ports, self.project_name)
            if error is True:
                click.secho(msg, fg='red')
                continue

            self._verbose(msg)


    def _run_services_post_scripts(self):
        """A service can have a .sh file that will be executed once it's started.
        Useful to override some actions of the classical /run.sh

        """

        if os.name == 'nt':
            click.secho('Could not run service post scripts under Windows', fg='red')
            return

        for service in self.main_config.get('services'):
            self._call_service_post_script(service)


    def _verbose(self, message: str):
        if self.context['VERBOSE'] is True:
            print(colored.green('[VERBOSE]') + ' {}'.format(message), file=sys.stderr)



import sys
import os
import shutil

from stakkr import package_utils
from setuptools.command.install import install


venv_dir = package_utils.get_venv_basedir()


def _post_install():
    required_dirs = [
        'conf',
        'conf/mysql-override',
        'conf/php-fpm-override',
        'conf/xhgui-override',
        'data',
        'home',
        'home/www-data',
        'home/www-data/bin',
        'logs',
        'plugins',
        'services',
        'www'
    ]
    for required_dir in required_dirs:
        _create_dir(required_dir)

    required_tpls = [
        '.env',
        'bash_completion',
        'conf/compose.ini.tpl',
        'conf/mysql-override/mysqld.cnf',
        'conf/php-fpm-override/example.conf',
        'conf/php-fpm-override/README',
        'conf/xhgui-override/config.php',
        'home/www-data/.bashrc'
    ]
    for required_tpl in required_tpls:
        _copy_file(required_tpl)


def _create_dir(dir_name: str):
    dir_name = venv_dir + '/' + dir_name.lstrip('/')
    if not os.path.isdir(dir_name):
        os.mkdir(dir_name)


def _copy_file(source_file: str):
    full_path = package_utils.get_file('tpls', source_file)
    dest_file = venv_dir + '/' + source_file
    if not os.path.isfile(dest_file):
        try:
            shutil.copy(full_path, dest_file)
        except Exception:
            msg = "Met an error trying to copy {} .. check that the file is there ...".format(full_path)
            print(msg, file=sys.stderr)



class StakkrPostInstall(install):
    def run(self):
        install.run(self)
        _post_install()
